#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <pwd.h>
#include "config.h"
#include "compose_response.h"
#include "log_stuff.h"


void log_closefd_err(struct in_addr *remote)
{
	syslog(LOG_ERR | LOG_DAEMON,
	       "X,X:%s: Failure to close file descriptor\n",
	       inet_ntoa(*remote));
}

int fetch_homedir(unsigned int uid, char *dir)
{
	struct passwd *b_pw;

	while ((b_pw = getpwent()) != NULL) {
		if (b_pw->pw_uid == uid) {
			strncpy(dir, b_pw->pw_dir, strlen(b_pw->pw_dir));
			return(0);
		}
	}
	return(1);	
}

int fetch_spoof_id(unsigned int uid, char *user)
{
	char homedir[MAX_SPOOFFILE_PATH];
	char spoof_file[14] = "/.ydentdspoof\0";
	FILE *fp;
	char spoof_id[MAX_USERNAME_LEN];
	char *b_user;

	if (fetch_homedir(uid, homedir) != 0) {
		/* ERRORHANDLER */
		return(1);
	}
	strncat(homedir, spoof_file, (MAX_SPOOFFILE_PATH - 14));
	if ((fp = fopen(homedir, "r")) == NULL) {
		/* ERRORHANDLER */
		return(1);
	}
	b_user = fgets(spoof_id, MAX_USERNAME_LEN, fp);
	strncpy(user, b_user, strlen(b_user));
	/* DEBUG */
	printf("I read a username of %s\n", user);
	return(0);
}

/* returns 0 on successful query response.
 * Returns 1 for cases where no response whatsoever was received
 * by the client. Any form of error responses return 0. TODO: Arrange
 * flexible return values to indicate whether:
 * 1. The client got a successful lookup
 * 2. The client got an error response (classified by nature of error)
 * 3. the client got squat.
 */
int service_request(int fd, unsigned long cadd, struct options *opts)
{
	unsigned int pos = 0;
	unsigned int poc = 0;
	unsigned int ewe_id = 0;
	char yewser[MAX_USERNAME_LEN] = "";
	char response[MAX_RESPONSE_SIZE] = "";
	char *fallback_response = "0 , 0 : ERROR : UNKNOWN-ERROR\n";
	char *esso = opts->opsys;

	struct in_addr buf_in_addr;
	buf_in_addr.s_addr = cadd;

	/* DEBUG */
	printf("in respond() allowspoof is %u\n", opts->allowuserspoof);

	switch (read_request(fd, &pos, &poc)) {
		case 0:	break; 
		case 1: /* technical failure to recv() */
			ylog(11, pos, poc, &buf_in_addr);
			if (answer(fd, fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
			}
			ylog(18, pos, poc, &buf_in_addr);
			return(0);
			// break;
		case 2:	/* recv()'d fine but couldn't make head or tail
			   of what we got */
			ylog(12, pos, poc, &buf_in_addr);
			if (answer(fd, fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
			}
			ylog(18, pos, poc, &buf_in_addr);
			return(0);
			// break;
		case 3:	/* one or both ports specified outside range */
			ylog(13, pos, poc, &buf_in_addr);
			if (compose_response(3, pos, poc, response,
						yewser, esso) != 0) {
				ylog(2, pos, poc, &buf_in_addr);
				if (answer(fd, fallback_response) != 0) {
					ylog(1, pos, poc, &buf_in_addr);
					if (close(fd) != 0) {
						log_closefd_err(&buf_in_addr);
					}
					return(0);
				}
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
				}
				ylog(18, pos, poc, &buf_in_addr);
				return(0);
			}
			if (answer(fd, response) != 0) {
				ylog(3, pos, poc, &buf_in_addr);
				if (answer(fd, fallback_response) != 0) {
					ylog(1, pos, poc, &buf_in_addr);
					if (close(fd) != 0) {
						log_closefd_err(&buf_in_addr);
					}
					return(0);
				}
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
				}
				ylog(18, pos, poc, &buf_in_addr);
				return(0);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
			}
			ylog(19, pos, poc, &buf_in_addr);
			return(0);
			// break;
	}
	if (find_ewe_id(pos, poc, cadd, &ewe_id) != 0) {
		ylog(20, pos, poc, &buf_in_addr);
		if (compose_response(4, pos, poc, response,
					yewser, esso) != 0) {
			ylog(4, pos, poc, &buf_in_addr);
			if (answer(fd,fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
					return(0);
				}
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
			}
			ylog(18, pos, poc, &buf_in_addr);
			return(0);
		}
		if (answer(fd, response) != 0 ) {
			ylog(5, pos, poc, &buf_in_addr);
			if (answer(fd, fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
				}
				return(0);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
			}
			ylog(18, pos, poc, &buf_in_addr);
			return(0);
		}
		if (close(fd) != 0) {
			log_closefd_err(&buf_in_addr);
			return(0);
		}
		ylog(21, pos, poc, &buf_in_addr);
		return(0);
	}
	ylog(14, pos, poc, &buf_in_addr);

	if (opts->allowuserspoof == 1) {
		if (fetch_spoof_id(ewe_id, yewser) == 0) {
			printf("I'm in respond.c and I got %s\n", yewser);
			ylog(24, pos, poc, &buf_in_addr);
			if (compose_response(6, pos, poc, response,
					     yewser, esso) != 0) {
				ylog(9, pos, poc, &buf_in_addr);
				if (answer(fd, fallback_response) != 0) {
					ylog(1, pos, poc, &buf_in_addr);
					if (close(fd) != 0) {
						log_closefd_err(&buf_in_addr);
					}
					return(0);
				}
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
					return(0);
				}
				return(0);
			}
			if (answer(fd, response) != 0) {
				ylog(10, pos, poc, &buf_in_addr);
				if (answer(fd, fallback_response) != 0) {
					ylog(1, pos, poc, &buf_in_addr);
					if (close(fd) != 0) {
						log_closefd_err(&buf_in_addr);
						return(0);
					}
				}
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
					return(0);
				}
				return(0);
			}
			return(0);
		}			
	}
	
	if (find_user_name(ewe_id, yewser) != 0) {
		ylog(22, pos, poc, &buf_in_addr);
		if (compose_response(5, pos, poc, response,
					yewser, esso) != 0) {
			ylog(8, pos, poc, &buf_in_addr);
			if (answer(fd, fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
				}
				return(0);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
				return(0);
			}
			ylog(18, pos, poc, &buf_in_addr);
			return(0);
		}
		if (answer(fd, response) != 0) {
			ylog(7, pos, poc, &buf_in_addr);
			if (answer(fd, fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
				}
				return(0);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
			}
			ylog(18, pos, poc, &buf_in_addr);
			return(0);
		}
		if (close(fd) != 0) {
			log_closefd_err(&buf_in_addr);
		}
		ylog(23, pos, poc, &buf_in_addr);
		return(0);
	}
	ylog(15, pos, poc, &buf_in_addr);
	
	/* Great. We have a port on the server and port on client, both valid.
	 * We also have successfully looked up a UID and an associated
	 * user name. Life is good. Let's tell the client so.
	 */
	if (compose_response(6, pos, poc, response, yewser, esso) != 0) {
		ylog(9, pos, poc, &buf_in_addr);
		if (compose_response(5, pos, poc, response,
					yewser, esso) != 0) {
			ylog(8, pos, poc, &buf_in_addr);
			if (answer(fd, fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
				}
				return(0);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
			}
			ylog(18, pos, poc, &buf_in_addr);
			return(0);
		}
		if (answer(fd, response) != 0) {
			ylog(7, pos, poc, &buf_in_addr);
			if (answer(fd, fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
				}
				return(0);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
			}
			ylog(18, pos, poc, &buf_in_addr);
			return(0);
		}
		if (close(fd) != 0) {
			log_closefd_err(&buf_in_addr);
		}
		ylog(23, pos, poc, &buf_in_addr);
		return(0);
	}
	ylog(16, pos, poc, &buf_in_addr);
	
	if (answer(fd, response) != 0 ) {
		ylog(10, pos, poc, &buf_in_addr);
		if (compose_response(5, pos, poc, response,
					yewser, esso) != 0) {
			ylog(8, pos, poc, &buf_in_addr);
			if (answer(fd, fallback_response) != 0) {
				ylog(1, pos, poc, &buf_in_addr);
				if (close(fd) != 0) {
					log_closefd_err(&buf_in_addr);
					return(0);
				}
				return(0);
			}
			if (close(fd) != 0) {
				log_closefd_err(&buf_in_addr);
				return(0);
			}
			return(0);
		}
		if (answer(fd, fallback_response) != 0) {
			ylog(1, pos, poc, &buf_in_addr);
			return(0);
		}
		if (close(fd) != 0) {
			log_closefd_err(&buf_in_addr);
		}
		ylog(18, pos, poc, &buf_in_addr);
		return(0);
	}
	ylog(17, pos, poc, &buf_in_addr);
	if (close(fd) != 0) {
		log_closefd_err(&buf_in_addr);
	}
	return(0);
}



/* returns:
 * 0 on success
 * 1 on failure 
 * 2 if recv succeeds but the pattern doesn't match "digit , digit"
 * 3 if the ports are outside tcp port range
 * */
int read_request(int fd, unsigned int *pos, unsigned int *poc)
{
	int numbytes = 0;
	char buf[MAXDATASIZE] = "";
	
	if ((numbytes = recv(fd, buf, (MAXDATASIZE - 1),0)) == -1) {
		// perror("recv");
		return(1);
	}

	buf[numbytes] = '\0';

	if (sscanf(buf, "%u , %u", pos, poc) != 2) {
		// perror("sscanf");
		return(2);
	}
	if ((valid_port(*pos) == 1) || (valid_port(*poc) == 1)) {
		return(3);
	}
	return 0;
}



/* returns 0 on success, 1 on error*/
int find_ewe_id(unsigned int pos, unsigned int poc,
		         unsigned long cadd, unsigned int *uid)
{
	FILE *tempfp;
	char *fgets_return;
	unsigned int result;
	char tmpbuf[MAX_RESPONSE_SIZE] = "";
	unsigned long b_cadd;
	unsigned int b_pos, b_poc;

	tempfp = fopen(TCP_LKP_FILE, "r");
	
	if (tempfp == NULL) {
		/* ERRORHANDLER */
		return(1);
	}
	
	if (fgets(tmpbuf, sizeof(tmpbuf), tempfp) == NULL ) {
		/* ERRORHANDLER */
		return(1);
	}

	while ((fgets_return = fgets(tmpbuf, sizeof(tmpbuf), tempfp))) {
		if (fgets_return == NULL) {
			/* ERRORHANDLER */
			return(1);
		}
		if (sscanf(tmpbuf,
		       "%*d: %*x:%x %lX:%x %*x %*X:%*X %*x:%*X %*x %u %*d %*d",
		       &b_pos, &b_cadd, &b_poc, &result) != 4) {
			/* ERRORHANDLER */
			return(1);
		}
		if (b_cadd == cadd &&
		    b_pos == pos &&
		    b_poc == poc) {
			*uid = result;
			if (fclose(tempfp) != 0) {
				/* ERRORHANDLER */
				return(1);
			}
			return 0;
		}
	}
	if (fclose(tempfp) != 0) {
		/* ERRORHANDLER */
		return(1);
	}
	return 1;
}

/* returns 0 on success, other on error
 * TODO adapt return values to be more meaningful on error
 * and adapt to respond to these different error conditions
 */
int find_user_name(unsigned int uid, char *eweser)
{
	FILE *fp;
	unsigned int tempuid;
	char tempname[MAX_USERNAME_LEN], tempbuf[MAX_PASSWDFILE_LEN];
	char *fgets_return;
	
	if ((fp = fopen(PASSWD_FILE, "r")) == NULL) {
		/* ERRORHANDLER */
		return(1);
	}
	while ((fgets_return = fgets(tempbuf, sizeof(tempbuf), fp))) {
		if (fgets_return == NULL) {
			/* ERRORHANDLER */
			if (fclose(fp) != 0) {
				/* ERRORHANDLER */
				exit(EXIT_FAILURE);
			}
			return(1);
		}
		if ((sscanf(tempbuf, "%[^:]:x:%u:%*s",
			    tempname, &tempuid)) != 2) {
			/* ERRORHANDLER */
			if (fclose(fp) != 0) {
				/* ERRORHANDLER */
				exit(EXIT_FAILURE);
			}
			return(1);
		}
		if (tempuid == uid) {
			if (fclose(fp) != 0) {
				/* ERRORHANDLER */
				exit(EXIT_FAILURE);
			}
			strncpy(eweser, tempname, strlen(tempname));
			return 0;
		}
	}
	if (fclose(fp) != 0) {
		/* ERRORHANDLER */
		exit(EXIT_FAILURE);
	}
	return 1;
}

/*    */
int answer(int fd, char *message)
{
	int len = strlen(message);
	if (send(fd, message, len, 0) != len) {
		return(1);
	}
	return(0);
}
	
int valid_port(unsigned int port) {
	if (port > 0) {
		if (port < 65536) {
			return 0;
		}
	}
	return 1;
}
/*
void log_closefd_err(struct in_addr *remote)
{
	syslog(LOG_ERR | LOG_DAEMON,
	       "X,X:%s: Failure to close file descriptor\n",
	       inet_ntoa(*remote));
}
*/
